##############################################################
# Drone Camera
# ver. 1.2.0 / 17.Oct.2020
# Innovare Studio Inc. / D.Tsukamoto
# ver. 1.2.0_keyboard / 16.Dec.2020 / leBluem
#############################################################

# also cursor keys are working for movement from normal freecam
KEYS = ['P', # INPUT.buttons.start       # 0    #    enable
		'M', # INPUT.buttons.l_thumb     # 1    #    altitude lock
		'N', # INPUT.buttons.r_thumb     # 2    #    stabilize
		'B', # INPUT.buttons.B           # 3    #    heading lock
		'I', # INPUT.buttons.A           # 4    #    autofocus
		'U', # INPUT.buttons.back        # 5    #    reset DOF
		'O', # INPUT.buttons.Y           # 6    #    follow car
		'J', # INPUT.buttons.X           # 7    #    heading mode
		'Z', # INPUT.buttons.povUp       # 8    #    DOF focus+
		'H', # INPUT.buttons.povDown     # 9    #    DOF focus-
		'T', # INPUT.buttons.povLeft     # 10   #    FOV+ zoom in
		'G', # INPUT.buttons.povRight    # 11   #    FOV- zoom out
		'R', # INPUT.l_trigger           # 12   #    move down
		'F', # INPUT.r_trigger           # 13   #    move up
		'C', # INPUT.r_thumb_x           # 14   #    rotate left
		'V', # INPUT.r_thumb_x           # 15   #    rotate right
		'Y', # INPUT.r_thumb_y           # 16   #    rotate forward
		'X', # INPUT.r_thumb_y           # 17   #    rotate backward
		'Q', # INPUT.buttons.L           # 18   #    roll left
		'E', # INPUT.buttons.R           # 19   #    roll right
		'W', # INPUT.l_thumb_y           # 20   #    forward
		'S', # INPUT.l_thumb_y           # 21   #    backward
		'A', # INPUT.l_thumb_x           # 22   #    left
		'D', # INPUT.l_thumb_x           # 23   #    right
	]

import ac, acsys
import os, sys, configparser, functools,  math, traceback, threading
import glob, json
from collections import OrderedDict
sys.path.insert(0, os.path.join(os.path.dirname(__file__)))
os.environ['PATH'] = os.environ['PATH'] + ";."
import ctypes

CONFIG_SPEED_ROT = 0.5
BUTTON_PRESS_WAIT = 0.5
AVAILABLE = False
ID_APP_MAIN = 0
ID_APP_CONFIG = 0
BUTTON_DELAY = 0

UI_FONT = "Segoe UI"
LABEL_BUTTONS = []

xinput = ctypes.windll.xinput9_1_0
XINPUT_DEVICES = []
XINPUT_MAX_DEVICES = 4

DOF_FOCUS_MAX = 500
DOF_AF_TARGET = 0

ENABLED_DRONE = False
ENABLED_STABILIZED = True
ENABLED_ALTITUDE_LOCK = True
ENABLED_FOLLOW_POS = True
ENABLED_FOLLOW_ANGLE = True
MODE_FOLLOW_ANGLE = 1

LAST_CAM_MODE = 0
FORCE_RESET = True
FOCUS = 0
NEAR_CAR = 0
CAM_POS = [0, 0, 0]
CAR_POS = [0, 0, 0]
CAR_HEADING = 0
CAR_YAW = 0

CAM_MOVED_LAST = [0, 0, 0]
CAM_REVERB = [0, 0, 0]
CAM_REVERB_DELAY = 0.0

LABEL_COMMON = {}
NATION_TEXTURES = {}
CAR_DETAILS = []

EVENTS = []

CONFIG_METHOD = ["MoveForward", "MoveRight", "MoveUpWorld", "RotateHeading", "RotatePitch", "RotateRoll"]

CONFIG = OrderedDict((
	["CONTROLLER", 0],
	["DEADZONE", 10],
	["SMOOTH", 40],
	["FOLLOW_SWITCH_SPEED", 30],
	["FOV_SPEED", 10],
	["DOF_SPEED", 50],
	["DOF_SPEED_AF", 30],
	["METHOD", OrderedDict()]
))
for method in CONFIG_METHOD:
	CONFIG["METHOD"][method] = {
		"SPEED_FOLLOW": 50,
		"SPEED": 50,
		"GAIN": 30,
		"INVERT": 0,
	}
	if method in ["MoveForward", "MoveRight", "MoveUpWorld"]:
		CONFIG["METHOD"][method]["SPEED_FOLLOW"] = 100
		CONFIG["METHOD"][method]["SPEED"] = 100
	if method in ["MoveForward", "MoveRight"]:
		CONFIG["METHOD"][method]["SPEED"] = 200

CONFIG_LABEL_NAME = {
	"CONTROLLER": "Controller ID",
	"DEADZONE": "Deadzone",
	"SMOOTH": "Smooth",
	"FOV_SPEED": "FOV speed",
	"DOF_SPEED": "DOF focus speed",
	"DOF_SPEED_AF": "Auto focus speed",
	"FOLLOW_SWITCH_SPEED": "Follow switch speed",
	"MoveForward": "Forward",
	"MoveRight": "Lateral",
	"MoveUpWorld": "Vertical",
	"RotateHeading": "Yaw",
	"RotatePitch": "Pitch",
	"RotateRoll": "Roll",
}
FILE_CONFIG = "apps/python/IS_DroneCam/config.ini"


class labelButton:
	def __init__(self, button, action, x, y, offset=0):
		self.button = button
		self.action = action
		self.offset = offset
		self.pos = [x, y]
		self.res = 0
		if len(XINPUT_DEVICES) > 0:
			self.res = ac.newTexture("apps/python/IS_DroneCam/res/icon_button_%s.png" % self.button)
		else:
			if self.button == "start":
				self.res = ac.newTexture("apps/python/IS_DroneCam/res/icon_button_keyboard.png")
		self.label = ac.addLabel(ID_APP_MAIN, self.action)
		ac.setPosition(self.label, self.pos[0] + 55 + self.offset, self.pos[1] + 5)

	def renderCallback(self):
		#global XINPUT_DEVICES
		#if len(XINPUT_DEVICES) > 0:
		if self.res != 0:
			ac.glQuadTextured(self.pos[0], self.pos[1], 64, 32, self.res)


def checkCSP():
	try:
		if not bool(ac.ext_patchVersion): return False
	except:
		return False
	return True


def configLoad():
	global CONFIG
	parser = configparser.ConfigParser()
	try:
		parser.read(FILE_CONFIG)
		for key in CONFIG:
			if key == "METHOD":
				for method in CONFIG_METHOD:
					CONFIG["METHOD"][method] = {
						"SPEED_FOLLOW": parser.getint(method, "SPEED_FOLLOW", fallback=CONFIG["METHOD"][method]["SPEED_FOLLOW"]),
						"SPEED": parser.getint(method, "SPEED", fallback=CONFIG["METHOD"][method]["SPEED"]),
						"GAIN": parser.getint(method, "GAIN", fallback=CONFIG["METHOD"][method]["GAIN"]),
						"INVERT": parser.getint(method, "INVERT", fallback=CONFIG["METHOD"][method]["INVERT"]),
					}
			else:
				CONFIG[key] = parser.getint("BASIC", key, fallback=CONFIG[key])
	except:
		return


def configSave():
	parser = configparser.ConfigParser()
	try:
		parser.add_section("BASIC")
		for key in CONFIG:
			if key == "METHOD":
				for method in CONFIG_METHOD:
					parser.add_section(method)
					parser.set(method, "SPEED_FOLLOW", str(int(CONFIG["METHOD"][method]["SPEED_FOLLOW"])))
					parser.set(method, "SPEED", str(int(CONFIG["METHOD"][method]["SPEED"])))
					parser.set(method, "GAIN", str(int(CONFIG["METHOD"][method]["GAIN"])))
					parser.set(method, "INVERT", str(int(CONFIG["METHOD"][method]["INVERT"])))
			else:
				parser.set("BASIC", key, str(int(CONFIG[key])))
		with open(FILE_CONFIG, "w") as data:
			parser.write(data)
	except:
		return


class XINPUT_BUTTONS(ctypes.LittleEndianStructure):
	_fields_ = [
		('povUp', ctypes.c_ushort, 1),
		('povDown', ctypes.c_ushort, 1),
		('povLeft', ctypes.c_ushort, 1),
		('povRight', ctypes.c_ushort, 1),
		('start', ctypes.c_ushort, 1),
		('back', ctypes.c_ushort, 1),
		('l_thumb', ctypes.c_ushort, 1),
		('r_thumb', ctypes.c_ushort, 1),
		('L', ctypes.c_ushort, 1),
		('R', ctypes.c_ushort, 1),
		('U1', ctypes.c_ushort, 1),
		('U2', ctypes.c_ushort, 1),
		('A', ctypes.c_ushort, 1),
		('B', ctypes.c_ushort, 1),
		('X', ctypes.c_ushort, 1),
		('Y', ctypes.c_ushort, 1),
	]


class XINPUT_STATE(ctypes.Structure):
	_fields_ = [
		('id', ctypes.c_ulong),
		('buttons', XINPUT_BUTTONS),
		('l_trigger', ctypes.c_ubyte),
		('r_trigger', ctypes.c_ubyte),
		('l_thumb_x', ctypes.c_short),
		('l_thumb_y', ctypes.c_short),
		('r_thumb_x', ctypes.c_short),
		('r_thumb_y', ctypes.c_short),
	]


def xinputEnumDevice():
	global XINPUT_DEVICES
	XINPUT_DEVICES = []
	for deviceId in range(XINPUT_MAX_DEVICES):
		if xinputGetState(deviceId) is not None:
			XINPUT_DEVICES.append(deviceId)


def xinputGetState(deviceId):
	state = XINPUT_STATE()
	res = xinput.XInputGetState(deviceId, ctypes.byref(state))
	if res == 0: return state


def n_axis(axis):
	DEADZONE = CONFIG["DEADZONE"]
	if axis > 0:
		if axis < 327.68 * DEADZONE:
			return 0
		else:
			return (axis - 327.68 * DEADZONE) / 32768.0
	else:
		if axis > -327.68 * DEADZONE:
			return 0
		else:
			return (axis + 327.68 * DEADZONE) / 32768.0


def n_triggers(t1, t2):
	DEADZONE = CONFIG["DEADZONE"]
	if t1 > 0:
		if t1 < 2.5 * DEADZONE:
			return 0
		else:
			return (t1 - 2.5 * DEADZONE) / 255.0
	if t2 > 0:
		if t2 < 2.5 * DEADZONE:
			return 0
		else:
			return -(t2 - 2.5 * DEADZONE) / 255.0
	return 0


def n_pov_x(left, right):
	ret = 1 if left else 0
	ret = -1 if right else ret
	return ret


def calcDistanceVec3(p1, p2):
	return math.sqrt(math.pow(p1[0] - p2[0], 2) + math.pow(p1[1] - p2[1], 2) + math.pow(p1[2] - p2[2], 2))


def calcDistanceVec2(p1, p2):
	return math.sqrt(math.pow(p1[0] - p2[0], 2) + math.pow(p1[2] - p2[2], 2))


def getCarHeading(index):
	v1 = ac.getCarState(index, acsys.CS.TyreContactPoint, acsys.WHEELS.FR)
	v2 = ac.getCarState(index, acsys.CS.TyreContactPoint, acsys.WHEELS.RR)
	return math.atan2(v2[2] - v1[2], v2[0] - v1[0]) * -1


def getNearCar():
	DISTANCE_LAST = 999999
	RET = 0
	for index in range(ac.getCarsCount()):
		if ac.isConnected(index):
			DISTANCE = calcDistanceVec3(CAM_POS, ac.getCarState(index, acsys.CS.WorldPosition))
			if DISTANCE < DISTANCE_LAST:
				DISTANCE_LAST = DISTANCE
				RET = index
	return RET



def onSetConfig(value, key):
	CONFIG[key] = value


def onSetMethodSpeedFollow(value, key):
	CONFIG["METHOD"][key]["SPEED_FOLLOW"] = value


def onSetMethodSpeed(value, key):
	CONFIG["METHOD"][key]["SPEED"] = value


def onSetMethodGain(value, key):
	CONFIG["METHOD"][key]["GAIN"] = value


def onSetMethodInvert(name, value, key):
	CONFIG["METHOD"][key]["INVERT"] = bool(value)


def acMain(version):
	global AVAILABLE
	AVAILABLE = checkCSP()
	if not AVAILABLE: return

	configLoad()
	xinputEnumDevice()

	global ID_APP_MAIN, ID_APP_CONFIG, LABEL_COMMON, NATION_TEXTURES, CAR_DETAILS
	global KEYS
	ID_APP_MAIN = ac.newApp("Drone Camera")
	ac.setTitle(ID_APP_MAIN, "")
	ac.setSize(ID_APP_MAIN, 550, 350)
	ac.drawBorder(ID_APP_MAIN, 0)
	ac.setIconPosition(ID_APP_MAIN, 0, -9000)
	ac.setBackgroundOpacity(ID_APP_MAIN, 0.5)
	ac.drawBackground(ID_APP_MAIN, 1)
	if len(XINPUT_DEVICES) > 0:
		ac.addRenderCallback(ID_APP_MAIN, acRender)
		LABEL_BUTTONS.append(labelButton("lrt", "Up/Down", 20, 57, 15))
		LABEL_BUTTONS.append(labelButton("lrb", "Roll", 190, 57, 15))
		LABEL_BUTTONS.append(labelButton("axisL", "Move", 310, 57, -5))
		LABEL_BUTTONS.append(labelButton("axisR", "Rotate", 420, 57, -5))
	else:
		ac.addRenderCallback(ID_APP_MAIN, acRender2)
		LABEL_BUTTONS.append(labelButton("axisL", KEYS[20]+"/"+KEYS[21]+"/"+KEYS[22]+"/"+KEYS[23]+" - Move", -40, 53, 15))
		LABEL_BUTTONS.append(labelButton("lrt", KEYS[12]+"/"+KEYS[13]+" - Up/Down", 100, 53, 15))
		LABEL_BUTTONS.append(labelButton("lrb", KEYS[18]+"/"+KEYS[19]+" - Roll", 270, 53, -5))
		LABEL_BUTTONS.append(labelButton("axisR", KEYS[14]+"/"+KEYS[15]+"/"+KEYS[16]+"/"+KEYS[17]+" - Rotate", 370, 53, -5))

	LABEL_BUTTONS.append(labelButton("start", "Drone Camera", 0, 13))
	LABEL_BUTTONS.append(labelButton("y", "Follow Car", 0, 105))
	LABEL_BUTTONS.append(labelButton("b", "Heading Lock", 0, 140))
	LABEL_BUTTONS.append(labelButton("x", "Heading Mode", 0, 175))
	LABEL_BUTTONS.append(labelButton("l", "Altitude Lock", 0, 230))
	LABEL_BUTTONS.append(labelButton("r", "Roll Stabilize", 0, 265))
	LABEL_BUTTONS.append(labelButton("povV", "FOV", 0, 300))
	LABEL_BUTTONS.append(labelButton("back", "DOF Mode", 265, 230))
	LABEL_BUTTONS.append(labelButton("a", "Auto Focus", 265, 265))
	LABEL_BUTTONS.append(labelButton("povH", "DOF Focus", 265, 300))

	LABEL_COMMON["FOLLOW"] = ac.addLabel(ID_APP_MAIN, "Follow Target:")
	ac.setPosition(LABEL_COMMON["FOLLOW"], 285, 110)

	LABEL_COMMON["FOLLOW_BACKGROUND"] = ac.addLabel(ID_APP_MAIN, "")
	ac.setPosition(LABEL_COMMON["FOLLOW_BACKGROUND"], 285, 135)
	ac.setSize(LABEL_COMMON["FOLLOW_BACKGROUND"], 240, 60)
	ac.setBackgroundColor(LABEL_COMMON["FOLLOW_BACKGROUND"], 0, 0, 0)
	ac.setBackgroundOpacity(LABEL_COMMON["FOLLOW_BACKGROUND"], 0.3)
	ac.drawBackground(LABEL_COMMON["FOLLOW_BACKGROUND"], 1)

	LABEL_COMMON["FOLLOW_PLAYER_NAME"] = ac.addLabel(ID_APP_MAIN, "")
	ac.setPosition(LABEL_COMMON["FOLLOW_PLAYER_NAME"], 285, 140)

	LABEL_COMMON["FOLLOW_CAR_NAME"] = ac.addLabel(ID_APP_MAIN, "")
	ac.setPosition(LABEL_COMMON["FOLLOW_CAR_NAME"], 285, 165)
	ac.setFontSize(LABEL_COMMON["FOLLOW_CAR_NAME"], 14)

	ID_APP_CONFIG = ac.newApp("Drone Camera Settings")
	ac.setTitle(ID_APP_CONFIG, "")
	ac.setSize(ID_APP_CONFIG, 725, 515)
	ac.drawBorder(ID_APP_CONFIG, 0)
	ac.setIconPosition(ID_APP_CONFIG, 0, -9000)
	ac.setBackgroundOpacity(ID_APP_CONFIG, 0.5)
	ac.drawBackground(ID_APP_CONFIG, 1)

	LABEL_COMMON["CONFIG_TITLE"] = ac.addLabel(ID_APP_CONFIG, " DRONE CAMERA SETTINGS")
	ac.setCustomFont(LABEL_COMMON["CONFIG_TITLE"], UI_FONT, 0, 1)
	ac.setFontColor(LABEL_COMMON["CONFIG_TITLE"], 0, 0, 0, 1)
	ac.setFontSize(LABEL_COMMON["CONFIG_TITLE"], 16)
	ac.setPosition(LABEL_COMMON["CONFIG_TITLE"], 20, 20)
	ac.setSize(LABEL_COMMON["CONFIG_TITLE"], 685, 23)
	ac.setBackgroundColor(LABEL_COMMON["CONFIG_TITLE"], 1, 1, 1)
	ac.setBackgroundOpacity(LABEL_COMMON["CONFIG_TITLE"], 1)
	ac.drawBackground(LABEL_COMMON["CONFIG_TITLE"], 1)

	X_OFFSET = 25
	Y_OFFSET = 75
	WIDTH = 175
	HEIGHT = 62
	ROW = 0
	COL = 0
	for OPTION in CONFIG:
		if OPTION == "METHOD": continue
		SPINNER = ac.addSpinner(ID_APP_CONFIG, CONFIG_LABEL_NAME[OPTION])
		ac.setPosition(SPINNER, X_OFFSET + (WIDTH * COL), Y_OFFSET + (HEIGHT * ROW))
		ac.setSize(SPINNER, 150, 25)
		if OPTION == "CONTROLLER":
			ac.setRange(SPINNER, 0, max(0, len(XINPUT_DEVICES) - 1))
		else:
			ac.setRange(SPINNER, 0, 100)
		ac.setValue(SPINNER, int(CONFIG[OPTION]))
		onSetValue = functools.partial(onSetConfig, key=OPTION)
		onSetValue.__name__ = "onSetConfig"
		EVENTS.append(onSetValue)
		ac.addOnValueChangeListener(SPINNER, onSetValue)
		COL += 1
		if COL >= 4:
			COL = 0
			ROW += 1

	Y_OFFSET = 215
	LABEL_WIDTH = 80
	WIDTH = 170
	HEIGHT = 50

	LABEL_HEAD_SPEED_FOLLOW = ac.addLabel(ID_APP_CONFIG, "Follow mode speed")
	ac.setFontAlignment(LABEL_HEAD_SPEED_FOLLOW, "center")
	ac.setPosition(LABEL_HEAD_SPEED_FOLLOW, X_OFFSET + LABEL_WIDTH, Y_OFFSET - 35)
	ac.setSize(LABEL_HEAD_SPEED_FOLLOW, WIDTH, 25)

	LABEL_HEAD_SPEED = ac.addLabel(ID_APP_CONFIG, "Track mode speed")
	ac.setFontAlignment(LABEL_HEAD_SPEED, "center")
	ac.setPosition(LABEL_HEAD_SPEED, X_OFFSET + LABEL_WIDTH + WIDTH + 20, Y_OFFSET - 35)
	ac.setSize(LABEL_HEAD_SPEED, WIDTH, 25)

	LABEL_HEAD_SPEED = ac.addLabel(ID_APP_CONFIG, "Gain")
	ac.setFontAlignment(LABEL_HEAD_SPEED, "center")
	ac.setPosition(LABEL_HEAD_SPEED, X_OFFSET + LABEL_WIDTH + (WIDTH * 2) + 40, Y_OFFSET - 35)
	ac.setSize(LABEL_HEAD_SPEED, WIDTH, 25)

	LABEL_HEAD_INVERT = ac.addLabel(ID_APP_CONFIG, "Invert")
	ac.setFontAlignment(LABEL_HEAD_INVERT, "center")
	ac.setPosition(LABEL_HEAD_INVERT, X_OFFSET + LABEL_WIDTH + (WIDTH * 3) + 40, Y_OFFSET - 35)
	ac.setSize(LABEL_HEAD_INVERT, 60, 25)

	ROW = 0
	for METHOD in CONFIG["METHOD"]:
		LABEL = ac.addLabel(ID_APP_CONFIG, CONFIG_LABEL_NAME[METHOD])
		ac.setPosition(LABEL, X_OFFSET, Y_OFFSET + (HEIGHT * ROW) + 2)
		ac.setSize(LABEL, LABEL_WIDTH, 25)

		BORDER = ac.addLabel(ID_APP_CONFIG, "")
		ac.setPosition(BORDER, X_OFFSET, Y_OFFSET + (HEIGHT * ROW) - 12)
		ac.setSize(BORDER, 675, 2)
		ac.setBackgroundColor(BORDER, 1, 1, 1)
		ac.setBackgroundOpacity(BORDER, 0.2)

		SPINNER_FOLLOW = ac.addSpinner(ID_APP_CONFIG, "")
		ac.setPosition(SPINNER_FOLLOW, X_OFFSET + LABEL_WIDTH, Y_OFFSET + (HEIGHT * ROW))
		ac.setSize(SPINNER_FOLLOW, WIDTH, 25)
		ac.setRange(SPINNER_FOLLOW, 0, 500)
		ac.setValue(SPINNER_FOLLOW, int(CONFIG["METHOD"][METHOD]["SPEED_FOLLOW"]))
		onSetValue = functools.partial(onSetMethodSpeedFollow, key=METHOD)
		onSetValue.__name__ = "onSetMethodSpeedFollow"
		EVENTS.append(onSetValue)
		ac.addOnValueChangeListener(SPINNER_FOLLOW, onSetValue)

		SPINNER = ac.addSpinner(ID_APP_CONFIG, "")
		ac.setPosition(SPINNER, X_OFFSET + LABEL_WIDTH + WIDTH + 20, Y_OFFSET + (HEIGHT * ROW))
		ac.setSize(SPINNER, WIDTH, 25)
		ac.setRange(SPINNER, 0, 500)
		ac.setValue(SPINNER, int(CONFIG["METHOD"][METHOD]["SPEED"]))
		onSetValue = functools.partial(onSetMethodSpeed, key=METHOD)
		onSetValue.__name__ = "onSetMethodSpeed"
		EVENTS.append(onSetValue)
		ac.addOnValueChangeListener(SPINNER, onSetValue)

		SPINNER_GAIN = ac.addSpinner(ID_APP_CONFIG, "")
		ac.setPosition(SPINNER_GAIN, X_OFFSET + LABEL_WIDTH + (WIDTH * 2) + 40, Y_OFFSET + (HEIGHT * ROW))
		ac.setSize(SPINNER_GAIN, WIDTH, 25)
		ac.setRange(SPINNER_GAIN, 0, 100)
		ac.setValue(SPINNER_GAIN, int(CONFIG["METHOD"][METHOD]["GAIN"]))
		onSetValue = functools.partial(onSetMethodGain, key=METHOD)
		onSetValue.__name__ = "onSetMethodGain"
		EVENTS.append(onSetValue)
		ac.addOnValueChangeListener(SPINNER_GAIN, onSetValue)

		INVERT = ac.addCheckBox(ID_APP_CONFIG, "")
		ac.setPosition(INVERT, X_OFFSET + LABEL_WIDTH + (WIDTH * 3) + 60, Y_OFFSET + (HEIGHT * ROW))
		ac.setValue(INVERT, int(CONFIG["METHOD"][METHOD]["INVERT"]))
		onSetValue = functools.partial(onSetMethodInvert, key=METHOD)
		onSetValue.__name__ = "onSetMethodInvert"
		EVENTS.append(onSetValue)
		ac.addOnCheckBoxChanged(INVERT, onSetValue)

		ROW += 1

	NATION_CODES = glob.glob("content/gui/NationFlags/*.png")
	for path in NATION_CODES:
		NATION = os.path.splitext(os.path.basename(path))[0]
		NATION_TEXTURES[NATION] = ac.newTexture("content/gui/NationFlags/%s.png" % NATION)

	for index in range(ac.getCarsCount()):
		CAR_ID = ac.getCarName(index)
		CAR_SKIN = ac.getCarSkin(index)
		CAR_JSON_PATH = "content/cars/%s/ui/ui_car.json" % CAR_ID
		try:
			with open(CAR_JSON_PATH, "r", encoding="utf-8") as FP:
				CAR_JSON = json.load(FP, strict=False)
		except:
			CAR_JSON = {"name": ""}
		DETAIL = {
			"NAME": CAR_JSON["name"],
			"RES_LIVERY": ac.newTexture("content/cars/%s/skins/%s/livery.png" % (CAR_ID, CAR_SKIN)),
		}
		CAR_DETAILS.append(DETAIL)


def acRender(delta):
	ac.glBegin(acsys.GL.Quads)
	for lb in LABEL_BUTTONS:
		ac.glColor4f(1, 1, 1, 1)
		if lb.button == "start":
			ac.setText(lb.label, "Drone Camera : %s" % ("Enabled" if ENABLED_DRONE else ("Disabled" if len(XINPUT_DEVICES) > 0 else "Controller not found")))
		else:
			ac.glColor4f(1, 1, 1, (1 if ENABLED_DRONE else 0.5))
			ac.setFontColor(lb.label, 1, 1, 1, (1 if ENABLED_DRONE else 0.5))

		if lb.button == "y":
			ac.setText(lb.label, "Follow Car : %s" % ("Enabled" if ENABLED_FOLLOW_POS else "Disabled"))
		if lb.button == "b":
			ac.setText(lb.label, "Heading Lock : %s" % ("Enabled" if ENABLED_FOLLOW_ANGLE else "Disabled"))
			ac.glColor4f(1, 1, 1, (1 if ENABLED_FOLLOW_POS and ENABLED_DRONE else 0.5))
			ac.setFontColor(lb.label, 1, 1, 1, (1 if ENABLED_FOLLOW_POS and ENABLED_DRONE else 0.5))
		if lb.button == "x":
			ac.setText(lb.label, "Heading Mode : %s" % ("Movement" if MODE_FOLLOW_ANGLE else "Yaw"))
			ac.glColor4f(1, 1, 1, (1 if ENABLED_FOLLOW_ANGLE and ENABLED_FOLLOW_POS and ENABLED_DRONE else 0.5))
			ac.setFontColor(lb.label, 1, 1, 1, (1 if ENABLED_FOLLOW_ANGLE and ENABLED_FOLLOW_POS and ENABLED_DRONE else 0.5))
		if lb.button == "l":
			ac.setText(lb.label, "Altitude Lock : %s" % ("Enabled" if ENABLED_ALTITUDE_LOCK else "Disabled"))
		if lb.button == "r":
			ac.setText(lb.label, "Stabilize : %s" % ("Enabled" if ENABLED_STABILIZED else "Disabled"))
		if lb.button == "back":
			ac.setText(lb.label, "DOF Mode : %s" % ("Enabled" if ac.ext_getCameraDofFactor() >= 1 else "Disabled"))
		if lb.button == "a":
			ac.glColor4f(1, 1, 1, (1 if ac.ext_getCameraDofFactor() >= 1 and ENABLED_DRONE else 0.5))
			ac.setFontColor(lb.label, 1, 1, 1, (1 if ac.ext_getCameraDofFactor() >= 1 and ENABLED_DRONE else 0.5))
		if lb.button == "povH":
			ac.setText(lb.label, "DOF Focus : {:.3f}".format(ac.ext_getCameraDofFocus()))
			ac.glColor4f(1, 1, 1, (1 if ac.ext_getCameraDofFactor() >= 1 and ENABLED_DRONE else 0.5))
			ac.setFontColor(lb.label, 1, 1, 1, (1 if ac.ext_getCameraDofFactor() >= 1 and ENABLED_DRONE else 0.5))
		if lb.button == "povV":
			ac.setText(lb.label, "FOV : {:.3f}".format(ac.ext_getCameraFov()))
		lb.renderCallback()

	ac.setBackgroundColor(LABEL_COMMON["FOLLOW_BACKGROUND"], 1 if ENABLED_FOLLOW_POS else 0, 0, 0)
	ac.setBackgroundOpacity(LABEL_COMMON["FOLLOW_BACKGROUND"], 0.3)

	ac.glColor4f(1, 1, 1, (1 if ENABLED_DRONE else 0.5))
	CAR_ID = FOCUS if ENABLED_FOLLOW_POS else NEAR_CAR
	NATION = ac.getDriverNationCode(CAR_ID)
	if NATION_TEXTURES.get(NATION, False):
		ac.glQuadTextured(295, 140, 20, 20, NATION_TEXTURES[NATION])
	ac.glQuadTextured(295, 165, 20, 20, CAR_DETAILS[CAR_ID]["RES_LIVERY"])
	ac.setText(LABEL_COMMON["FOLLOW"], "Follow Target (%s) :" % ("Locked" if ENABLED_FOLLOW_POS else "Near"))
	ac.setFontColor(LABEL_COMMON["FOLLOW"], 1, 1, 1, (1 if ENABLED_DRONE else 0.5))
	ac.setText(LABEL_COMMON["FOLLOW_PLAYER_NAME"], "         %s" % ac.getDriverName(CAR_ID))
	ac.setFontColor(LABEL_COMMON["FOLLOW_PLAYER_NAME"], 1, 1, 1, (1 if ENABLED_DRONE else 0.5))
	ac.setText(LABEL_COMMON["FOLLOW_CAR_NAME"], "          %s" % CAR_DETAILS[CAR_ID]["NAME"])
	ac.setFontColor(LABEL_COMMON["FOLLOW_CAR_NAME"], 1, 1, 1, (1 if ENABLED_DRONE else 0.5))

	ac.glEnd()
	ac.setBackgroundOpacity(ID_APP_MAIN, 0.5)
	ac.drawBackground(ID_APP_MAIN, 1)

###
def acRender2(delta):
	global KEYS
	ac.glBegin(acsys.GL.Quads)
	for lb in LABEL_BUTTONS:
		ac.glColor4f(1, 1, 1, 1)
		if lb.button == "start":
			#ac.setText(lb.label, "Drone Camera : %s" % ("Enabled" if ENABLED_DRONE else ("Disabled" if len(XINPUT_DEVICES) > 0 else "Controller not found")))
			ac.setText(lb.label, KEYS[ 0]+" - Drone Camera : %s" % ("Enabled" if ENABLED_DRONE else "Disabled"))
		else:
			ac.glColor4f(1, 1, 1, (1 if ENABLED_DRONE else 0.5))
			ac.setFontColor(lb.label, 1, 1, 1, (1 if ENABLED_DRONE else 0.5))

		if lb.button == "y":
			ac.setText(lb.label, KEYS[ 6]+" - Follow Car : %s" % ("Enabled" if ENABLED_FOLLOW_POS else "Disabled"))
		if lb.button == "b":
			# ac.setText(lb.label, "Heading Lock : %s" % ("Enabled" if ENABLED_FOLLOW_ANGLE else "Disabled"))
			ac.setText(lb.label, KEYS[ 3]+" - Heading Lock : %s" % ("Enabled" if ENABLED_FOLLOW_ANGLE else "Disabled"))
			ac.glColor4f(1, 1, 1, (1 if ENABLED_FOLLOW_POS and ENABLED_DRONE else 0.5))
			ac.setFontColor(lb.label, 1, 1, 1, (1 if ENABLED_FOLLOW_POS and ENABLED_DRONE else 0.5))
		if lb.button == "x":
			# ac.setText(lb.label, "Heading Mode : %s" % ("Movement" if MODE_FOLLOW_ANGLE else "Yaw"))
			ac.setText(lb.label, KEYS[ 7]+" - Heading Mode : %s" % ("Movement" if MODE_FOLLOW_ANGLE else "Yaw"))
			ac.glColor4f(1, 1, 1, (1 if ENABLED_FOLLOW_ANGLE and ENABLED_FOLLOW_POS and ENABLED_DRONE else 0.5))
			ac.setFontColor(lb.label, 1, 1, 1, (1 if ENABLED_FOLLOW_ANGLE and ENABLED_FOLLOW_POS and ENABLED_DRONE else 0.5))
		if lb.button == "l":
			#ac.setText(lb.label, "Altitude Lock : %s" % ("Enabled" if ENABLED_ALTITUDE_LOCK else "Disabled"))
			ac.setText(lb.label, KEYS[ 1]+" - Altitude Lock : %s" % ("Enabled" if ENABLED_ALTITUDE_LOCK else "Disabled"))
		if lb.button == "r":
			#ac.setText(lb.label, "Stabilize : %s" % ("Enabled" if ENABLED_STABILIZED else "Disabled"))
			ac.setText(lb.label, KEYS[ 2]+" - Stabilize : %s" % ("Enabled" if ENABLED_STABILIZED else "Disabled"))
		if lb.button == "back":
			#ac.setText(lb.label, "DOF Mode : %s" % ("Enabled" if ac.ext_getCameraDofFactor() >= 1 else "Disabled"))
			ac.setText(lb.label, KEYS[ 5]+" - DOF Mode : %s" % ("Enabled" if ac.ext_getCameraDofFactor() >= 1 else "Disabled"))
		if lb.button == "a":
			ac.setText(lb.label, KEYS[ 4]+" - Auto Focus : %s" % ("Enabled" if ac.ext_getCameraDofFactor() >= 1 else "Disabled"))
			ac.glColor4f(1, 1, 1, (1 if ac.ext_getCameraDofFactor() >= 1 and ENABLED_DRONE else 0.5))
			ac.setFontColor(lb.label, 1, 1, 1, (1 if ac.ext_getCameraDofFactor() >= 1 and ENABLED_DRONE else 0.5))
		if lb.button == "povH":
			ac.setText(lb.label, KEYS[ 8]+"/"+KEYS[ 9]+" - DOF Focus : {:.3f}".format(ac.ext_getCameraDofFocus()))
			ac.glColor4f(1, 1, 1, (1 if ac.ext_getCameraDofFactor() >= 1 and ENABLED_DRONE else 0.5))
			ac.setFontColor(lb.label, 1, 1, 1, (1 if ac.ext_getCameraDofFactor() >= 1 and ENABLED_DRONE else 0.5))
		if lb.button == "povV":
			ac.setText(lb.label, KEYS[10]+"/"+KEYS[11]+" - FOV : {:.3f}\n(shift for 'dolly' zoom)".format(ac.ext_getCameraFov()))
		lb.renderCallback()

	ac.setBackgroundColor(LABEL_COMMON["FOLLOW_BACKGROUND"], 1 if ENABLED_FOLLOW_POS else 0, 0, 0)
	ac.setBackgroundOpacity(LABEL_COMMON["FOLLOW_BACKGROUND"], 0.3)

	ac.glColor4f(1, 1, 1, (1 if ENABLED_DRONE else 0.5))
	CAR_ID = FOCUS if ENABLED_FOLLOW_POS else NEAR_CAR
	NATION = ac.getDriverNationCode(CAR_ID)
	if NATION_TEXTURES.get(NATION, False):
		ac.glQuadTextured(295, 140, 20, 20, NATION_TEXTURES[NATION])
	ac.glQuadTextured(295, 165, 20, 20, CAR_DETAILS[CAR_ID]["RES_LIVERY"])
	ac.setText(LABEL_COMMON["FOLLOW"], "Follow Target (%s) :" % ("Locked" if ENABLED_FOLLOW_POS else "Near"))
	ac.setFontColor(LABEL_COMMON["FOLLOW"], 1, 1, 1, (1 if ENABLED_DRONE else 0.5))
	ac.setText(LABEL_COMMON["FOLLOW_PLAYER_NAME"], "         %s" % ac.getDriverName(CAR_ID))
	ac.setFontColor(LABEL_COMMON["FOLLOW_PLAYER_NAME"], 1, 1, 1, (1 if ENABLED_DRONE else 0.5))
	ac.setText(LABEL_COMMON["FOLLOW_CAR_NAME"], "          %s" % CAR_DETAILS[CAR_ID]["NAME"])
	ac.setFontColor(LABEL_COMMON["FOLLOW_CAR_NAME"], 1, 1, 1, (1 if ENABLED_DRONE else 0.5))

	ac.glEnd()
	ac.setBackgroundOpacity(ID_APP_MAIN, 0.5)
	ac.drawBackground(ID_APP_MAIN, 1)


BUTTONS_DELAY = {
	"start": 0,
	"back": 0,
	"A": 0,
	"B": 0,
	"X": 0,
	"Y": 0,
	"l_thumb": 0,
	"r_thumb": 0,
}


def acUpdate(delta):
	try:
		if not AVAILABLE: return
		# if len(XINPUT_DEVICES) == 0: return

		global ENABLED_DRONE, ENABLED_FOLLOW_POS, ENABLED_FOLLOW_ANGLE, MODE_FOLLOW_ANGLE, ENABLED_STABILIZED, ENABLED_ALTITUDE_LOCK
		global BUTTONS_DELAY, CAM_REVERB_DELAY
		global LAST_CAM_MODE, CAM_MOVED_LAST, LAST_SPEEDS
		global CAM_REVERB, DOF_AF_TARGET
		global KEYS, ID_APP_MAIN

		for BUTTON in BUTTONS_DELAY:
			BUTTONS_DELAY[BUTTON] += delta
		CAM_REVERB_DELAY += delta

		if len(XINPUT_DEVICES) > 0:
			INPUT = xinputGetState(XINPUT_DEVICES[CONFIG["CONTROLLER"]])
		####################################################################################
		else:
			INPUT = XINPUT_STATE() # create fake data
			if     ac.ext_isButtonPressed(KEYS[ 0])  : INPUT.buttons.start   = True
			if ENABLED_DRONE:
				if ac.ext_isButtonPressed(KEYS[ 1])  : INPUT.buttons.l_thumb = True
				if ac.ext_isButtonPressed(KEYS[ 2])  : INPUT.buttons.r_thumb = True
				if ac.ext_isButtonPressed(KEYS[ 3])  : INPUT.buttons.B       = True
				if ac.ext_isButtonPressed(KEYS[ 4])  : INPUT.buttons.A       = True
				if ac.ext_isButtonPressed(KEYS[ 5])  : INPUT.buttons.back    = True
				if ac.ext_isButtonPressed(KEYS[ 6])  : INPUT.buttons.Y       = True
				if ac.ext_isButtonPressed(KEYS[ 7])  : INPUT.buttons.X       = True
				if ac.ext_isButtonPressed(KEYS[ 8])  : INPUT.buttons.povUp   = True
				if ac.ext_isButtonPressed(KEYS[ 9])  : INPUT.buttons.povDown = True
				if ac.ext_isButtonPressed(KEYS[10])  : INPUT.buttons.povLeft = True
				if ac.ext_isButtonPressed(KEYS[11])  : INPUT.buttons.povRight= True
				if ac.ext_isButtonPressed(KEYS[12])  : INPUT.l_trigger       = -(CONFIG["DEADZONE"]+1)
				if ac.ext_isButtonPressed(KEYS[13])  : INPUT.r_trigger       = -(CONFIG["DEADZONE"]+1)
				if ac.ext_isButtonPressed(KEYS[14])  : INPUT.r_thumb_x       = 32768+CONFIG["DEADZONE"]*328
				if ac.ext_isButtonPressed(KEYS[15])  : INPUT.r_thumb_x       = 32768-CONFIG["DEADZONE"]*328
				if ac.ext_isButtonPressed(KEYS[16])  : INPUT.r_thumb_y       = 32768+CONFIG["DEADZONE"]*328
				if ac.ext_isButtonPressed(KEYS[17])  : INPUT.r_thumb_y       = 32768-CONFIG["DEADZONE"]*328
				if ac.ext_isButtonPressed(KEYS[18])  : INPUT.buttons.L       = True
				if ac.ext_isButtonPressed(KEYS[19])  : INPUT.buttons.R       = True
				if ac.ext_isButtonPressed(KEYS[20])  : INPUT.l_thumb_y       = 32768-CONFIG["DEADZONE"]*328
				if ac.ext_isButtonPressed(KEYS[21])  : INPUT.l_thumb_y       = 32768+CONFIG["DEADZONE"]*328
				if ac.ext_isButtonPressed(KEYS[22])  : INPUT.l_thumb_x       = 32768+CONFIG["DEADZONE"]*328
				if ac.ext_isButtonPressed(KEYS[23])  : INPUT.l_thumb_x       = 32768-CONFIG["DEADZONE"]*328
		####################################################################################

		global FORCE_RESET, CAM_POS, CAR_POS, CAR_HEADING, CAR_YAW, NEAR_CAR, FOCUS
		CAM_POS = list(ac.ext_getCameraPosition())
		CAM_POS_BEFORE = list(ac.ext_getCameraPosition())
		DOF_FACTOR = ac.ext_getCameraDofFactor()
		CAR_POS_LAST = CAR_POS
		CAR_POS = list(ac.getCarState(FOCUS, acsys.CS.WorldPosition))

		if ENABLED_DRONE and ac.getCameraMode() != acsys.CM.Free:
			ENABLED_DRONE = False

		FOCUS = ac.getFocusedCar()
		NEAR_CAR = FOCUS
		if ENABLED_DRONE and not ENABLED_FOLLOW_POS:
			NEAR_CAR = getNearCar()

		if INPUT.buttons.start and BUTTONS_DELAY["start"] > BUTTON_PRESS_WAIT:
			ENABLED_DRONE = not ENABLED_DRONE
			if ENABLED_DRONE:
				ac.setCameraMode(acsys.CM.Free)
				FORCE_RESET = True
			else:
				ac.setCameraMode(LAST_CAM_MODE)
			BUTTONS_DELAY["start"] = 0

		if INPUT.buttons.back and BUTTONS_DELAY["back"] > BUTTON_PRESS_WAIT:
			DOF_FACTOR = 1 if DOF_FACTOR <= 0 else 0
			BUTTONS_DELAY["back"] = 0

		if INPUT.buttons.Y and BUTTONS_DELAY["Y"] > BUTTON_PRESS_WAIT:
			ENABLED_FOLLOW_POS = not ENABLED_FOLLOW_POS
			if ENABLED_FOLLOW_POS:
				FOCUS = NEAR_CAR
				ac.focusCar(FOCUS)
				CAR_POS = list(ac.getCarState(FOCUS, acsys.CS.WorldPosition))
				CAR_POS_LAST = CAR_POS
				FORCE_RESET = True
			BUTTONS_DELAY["Y"] = 0
			CAM_REVERB_DELAY = 0
			CAM_REVERB = CAM_MOVED_LAST
			LAST_SPEEDS = {
				"MoveForward": 0,
				"MoveRight": 0,
				"MoveUpWorld": 0,
				"RotateHeading": 0,
				"RotatePitch": 0,
				"RotateRoll": 0,
			}

		if INPUT.buttons.B and BUTTONS_DELAY["B"] > BUTTON_PRESS_WAIT:
			ENABLED_FOLLOW_ANGLE = not ENABLED_FOLLOW_ANGLE
			if ENABLED_FOLLOW_ANGLE:
				FORCE_RESET = True
			BUTTONS_DELAY["B"] = 0

		if INPUT.buttons.X and BUTTONS_DELAY["X"] > BUTTON_PRESS_WAIT:
			MODE_FOLLOW_ANGLE = 0 if MODE_FOLLOW_ANGLE == 1 else 1
			FORCE_RESET = True
			BUTTONS_DELAY["X"] = 0

		if INPUT.buttons.l_thumb and BUTTONS_DELAY["l_thumb"] > BUTTON_PRESS_WAIT:
			ENABLED_ALTITUDE_LOCK = not ENABLED_ALTITUDE_LOCK
			BUTTONS_DELAY["l_thumb"] = 0

		if INPUT.buttons.r_thumb and BUTTONS_DELAY["r_thumb"] > BUTTON_PRESS_WAIT:
			ENABLED_STABILIZED = not ENABLED_STABILIZED
			BUTTONS_DELAY["r_thumb"] = 0

		if INPUT.buttons.A:
			DOF_AF_TARGET = max(0.5, calcDistanceVec2(CAM_POS, CAR_POS) - 1.0)

		if FORCE_RESET:
			FORCE_RESET = False
			CAR_HEADING = getCarHeading(FOCUS)
			CAR_VELOCITY = ac.getCarState(FOCUS, acsys.CS.Velocity)
			CAR_YAW = math.atan2(CAR_VELOCITY[2], CAR_VELOCITY[0]) * -1

		if ENABLED_DRONE:
			CAM_REVERB_DURATION = max(0.001, CONFIG["FOLLOW_SWITCH_SPEED"] / 10)
			if not ENABLED_FOLLOW_POS and CAM_REVERB_DELAY < CAM_REVERB_DURATION:
				for i in range(0, 3):
					CAM_POS[i] += CAM_MOVED_LAST[i] * (max(0.0, math.pow((CAM_REVERB_DURATION - CAM_REVERB_DELAY) / CAM_REVERB_DURATION, 0.2)))

			DOF_FOCUS = ac.ext_getCameraDofFocus()
			FOV = ac.ext_getCameraFov()
			ac.ext_setCameraDofFactor(DOF_FACTOR)
			DOF_FOCUS_CHANGE = n_pov_x(INPUT.buttons.povUp, INPUT.buttons.povDown)
			if DOF_FOCUS <= 0 and DOF_FOCUS_CHANGE < 0:
				ac.ext_setCameraDofFocus(0)
				DOF_AF_TARGET = 0
			elif DOF_FOCUS >= DOF_FOCUS_MAX and DOF_FOCUS_CHANGE > 0:
				DOF_AF_TARGET = DOF_FOCUS_MAX
				ac.ext_setCameraDofFocus(DOF_FOCUS_MAX)
			elif DOF_FOCUS_CHANGE != 0:
				ac.ext_setCameraDofFocus(DOF_FOCUS + (DOF_FOCUS_CHANGE * (max((DOF_FOCUS * (CONFIG["DOF_SPEED"] / 1000)), 0.001))))
				DOF_AF_TARGET = ac.ext_getCameraDofFocus()
			elif DOF_AF_TARGET != DOF_FOCUS:
				DOF_FOCUS += (DOF_AF_TARGET - DOF_FOCUS) * (math.pow(CONFIG["DOF_SPEED_AF"] * 0.1, 1.5) / 100)
				ac.ext_setCameraDofFocus(DOF_FOCUS)

			FOV_CHANGE = n_pov_x(INPUT.buttons.povLeft, INPUT.buttons.povRight)
			if FOV <= 1 and FOV_CHANGE < 0: FOV_CHANGE = 0
			fFOV = FOV_CHANGE * (CONFIG["FOV_SPEED"] / 100)
			ac.ext_setCameraFov( FOV + fFOV )

			if ENABLED_FOLLOW_POS:

				for i in range(3):
					CAM_POS[i] += (CAR_POS[i] - CAR_POS_LAST[i]) * (min(1.0, math.pow(CAM_REVERB_DELAY / CAM_REVERB_DURATION, 0.2)))

				if ENABLED_FOLLOW_ANGLE:
					DISTANCE = calcDistanceVec2(CAM_POS, CAR_POS)
					ANGLE = 0

					CAR_HEADING_LAST = CAR_HEADING
					CAR_HEADING = getCarHeading(FOCUS)
					CAR_VELOCITY = ac.getCarState(FOCUS, acsys.CS.Velocity)
					CAR_YAW_LAST = CAR_YAW
					CAR_YAW = math.atan2(CAR_VELOCITY[2], CAR_VELOCITY[0]) * -1

					if MODE_FOLLOW_ANGLE == 0 or ac.getCarState(FOCUS, acsys.CS.SpeedKMH) < 2:
						ANGLE = CAR_HEADING_LAST - CAR_HEADING
					elif MODE_FOLLOW_ANGLE == 1:
						ANGLE = CAR_YAW_LAST - CAR_YAW

					CAM_DIRECTION = math.atan2(CAM_POS[2] - CAR_POS[2], CAM_POS[0] - CAR_POS[0]) * -1

					####################################################################################
					if ac.ext_isShiftPressed() and (INPUT.buttons.povLeft or INPUT.buttons.povRight):
						# do a wonky dolly zoom with fov +/- and shift held down
						CAM_POS[0] = CAR_POS[0] + math.cos(CAM_DIRECTION - ANGLE) * (DISTANCE-fFOV/CONFIG["FOV_SPEED"]/4)
						#CAM_POS[1] = CAM_POS[1] + math.sin(CAM_DIRECTION - ANGLE) * (fFOV/CONFIG["FOV_SPEED"])
						CAM_POS[2] = CAR_POS[2] - math.sin(CAM_DIRECTION - ANGLE) * (DISTANCE-fFOV/CONFIG["FOV_SPEED"]/4)
					else:
					####################################################################################
						CAM_POS[0] = CAR_POS[0] + math.cos(CAM_DIRECTION - ANGLE) * DISTANCE
						CAM_POS[2] = CAR_POS[2] - math.sin(CAM_DIRECTION - ANGLE) * DISTANCE

					#CAM_YAW = ac.ext_getCameraYawRad()
					CAM_PITCH = ac.ext_getCameraPitchRad()
					CAM_ROLL = ac.ext_getCameraRollRad()
					ac.freeCameraRotateHeading(-ANGLE)
					#DIFF = CAM_YAW - ac.ext_getCameraYawRad()
					#ac.freeCameraRotateHeading(ANGLE)
					#ac.freeCameraRotateHeading(-ANGLE * (ANGLE / DIFF))
					ac.freeCameraRotatePitch(CAM_PITCH - ac.ext_getCameraPitchRad())
					ac.freeCameraRotateRoll(ac.ext_getCameraRollRad() - CAM_ROLL)

			ac.ext_setCameraPosition(*CAM_POS)

			cameraMove("MoveForward"  , delta,  n_axis(INPUT.l_thumb_y))
			cameraMove("MoveRight"    , delta,  n_axis(INPUT.l_thumb_x))
			cameraMove("MoveUpWorld"  , delta, -n_triggers(INPUT.l_trigger, INPUT.r_trigger))
			cameraMove("RotateHeading", delta, -n_axis(INPUT.r_thumb_x), rotate=CONFIG_SPEED_ROT)
			cameraMove("RotatePitch"  , delta,  n_axis(INPUT.r_thumb_y), rotate=CONFIG_SPEED_ROT)
			cameraMove("RotateRoll"   , delta,  n_pov_x(INPUT.buttons.L, INPUT.buttons.R), rotate=CONFIG_SPEED_ROT)

			CAM_POS_AFTER = list(ac.ext_getCameraPosition())
			for i in range(3):
				CAM_MOVED_LAST[i] = CAM_POS_AFTER[i] - CAM_POS_BEFORE[i]
		else:
			LAST_CAM_MODE = ac.getCameraMode()
	except:
		ac.log("IS_DRONECam error:" + traceback.format_exc())


LAST_SPEEDS = {
	"MoveForward": 0,
	"MoveRight": 0,
	"MoveUpWorld": 0,
	"RotateHeading": 0,
	"RotatePitch": 0,
	"RotateRoll": 0,
}


def cameraMove(type, delta, input, rotate=1.0):
	global LAST_SPEEDS
	smooth = 10 / ((CONFIG["SMOOTH"] / 10) + 0.1)
	gain = 1 + (CONFIG["METHOD"][type]["GAIN"] / 50)
	lsp = math.pow(abs(input), gain) * (-1 if input < 0 else 1)
	if smooth > 0:
		po = (2.718 ** (-smooth * delta))
		LAST_SPEEDS[type] = po * LAST_SPEEDS[type] + (1 - po) * lsp
		lsp = LAST_SPEEDS[type]
	if type == "RotateHeading" and ENABLED_STABILIZED:
		lsp *= 1.0 - abs(ac.ext_getCameraPitchRad() / (math.pi / 2))
	invert = -1 if CONFIG["METHOD"][type]["INVERT"] == 1 else 1
	CAM_PITCH = ac.ext_getCameraPitchRad()
	CAM_ROLL = ac.ext_getCameraRollRad()
	CAM_POS_OLD = ac.ext_getCameraPosition()
	SPEED = CONFIG["METHOD"][type]["SPEED_FOLLOW"] if ENABLED_FOLLOW_POS else CONFIG["METHOD"][type]["SPEED"]
	eval('ac.freeCamera' + type)((SPEED / 10) * delta * rotate * invert * lsp)
	if type == "RotateHeading" and ENABLED_STABILIZED:
		ac.freeCameraRotatePitch(CAM_PITCH - ac.ext_getCameraPitchRad())
		ac.freeCameraRotateRoll(ac.ext_getCameraRollRad() - CAM_ROLL)
	CAM_POS_NEW = ac.ext_getCameraPosition()
	if ENABLED_ALTITUDE_LOCK and type != "MoveUpWorld":
		ac.ext_setCameraPosition(CAM_POS_NEW[0], CAM_POS_OLD[1], CAM_POS_NEW[2])


def easing(time, base, target, duration):
	return target * (0.5 - math.cos(time / duration * math.pi) / 2) + base


def acShutdown():
	if not AVAILABLE: return
	configSave()
